package tdp024.todo.data.db.facade;

import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.LockModeType;
import javax.persistence.Query;
import javax.persistence.QueryTimeoutException;
import javax.persistence.RollbackException;
import tdp024.todo.data.api.entity.Todo;
import tdp024.todo.data.api.exception.TodoEntityNotFoundException;
import tdp024.todo.data.api.exception.TodoInputParameterException;
import tdp024.todo.data.api.exception.TodoServiceConfigurationException;
import tdp024.todo.data.db.entity.TodoDB;
import tdp024.todo.data.api.facade.TodoEntityFacade;
import tdp024.todo.data.db.util.EMF;
import tdp024.util.logger.TodoLogger;
import tdp024.util.logger.TodoLoggerImpl;

public class TodoEntityFacadeDB implements TodoEntityFacade {

    private static final TodoLogger todoLogger = new TodoLoggerImpl();

    @Override
    public long create(String title, String body)
            throws
            TodoInputParameterException,
            TodoServiceConfigurationException {

        if (title == null || body == null || title.trim().equals("") || body.trim().equals("")) {
            /*
             * We wont even bother doing the rest of the method if title or body is null.
             * 
             * We do a DEBUG log, and then throw a new exception that we let the caller handle
             * (in this case the logic layer)
             */
            todoLogger.log(TodoLogger.TodoLoggerLevel.DEBUG, "Title or body was null", "Title or body was null");
            throw new TodoInputParameterException("Neither title nor body can be null");
        }

        EntityManager em = EMF.getEntityManager();


        /*
         * What exceptions do we need to handle?
         * 
         * Look at the java doc for:
         *  EntityManager
         *  and
         *  EntityTransaction
         * 
         *  Look at the methods we use from these classes, what
         *  exceptions might occur? Do we need to check for all of them?
         *  Are some impossible to happen, so we can ignore them?
         */
        try {

            /*
             * This could throw IllegalStateException in case we try
             * to begin the transaction twice. That is quite unlikely
             * as we can control this ourselves, so we ignore this exception.
             */
            em.getTransaction().begin();

            /* 
             * This wont throw any exceptions
             */
            Todo todo = new TodoDB();
            todo.setTitle(title);
            todo.setBody(body);

            /*
             * In the java doc we have:
             * 
             * EntityExistsException - if the entity already exists. 
             * IllegalArgumentException - if the instance is not an entity
             * TransactionRequiredException - if invoked on a container-managed entity manager of type PersistenceContextType.TRANSACTION and there is no transaction
             * 
             * Since we actually created a new Todo object, we know that it can not exist 
             * in the database, so EntityExistsException will not occur.
             * 
             * We know that Todo is an entity, so IllegalArgumentException wont happen.
             * 
             * TransactionRequiredException, we have started the transaction so we know that this wont happen
             * 
             * This means that all the exceptions that might occur from "persist" are within our control, so 
             * we do not HAVE to check for them (but we could if we feel that we do not trust ourselves).
             * 
             */
            em.persist(todo);

            /*
             * From the java doc
             * IllegalStateException - if isActive() is false
             * RollbackException - if the commit fails
             * 
             * The first exception (IllegalStateException) we can control, we know that we have started the transaction earlier.
             * 
             * The second exception (RollbackException) is the only one that we can not control. The reasons for this exceptions
             * could be many, e.g. the network is down, the database is corrupted, etc etc. In any case this exception is such
             * that we can not "code our way out of it", and the calling code (in this case the logic layer) can not solve this
             * either. This exception should be checked for, and if it happens we should rollback, log and throw a TodoServiceConfigurationException
             * 
             */
            em.getTransaction().commit();

            return todo.getId();

        } catch (RollbackException e) {

            todoLogger.log(e);

            /* We have moved the rollback into the catch of rollback exception */
            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }

            throw new TodoServiceConfigurationException("Creating the Todo failed due to service errors. Please contact your database administrator.");

        } finally {

            em.close();
        }

    }

    @Override
    public void setActive(long id, boolean active)
            throws
            TodoEntityNotFoundException,
            TodoServiceConfigurationException {

        EntityManager em = EMF.getEntityManager();

        try {

            /* We handled this one in create */
            em.getTransaction().begin();

            /*
             * From java doc
             * IllegalArgumentException - if the first argument does not denote an entity type or the second argument is is not a valid type for that entity’s primary key or is null
             * 
             * We can control this, by statically typing id to be a long as input parameter and also hard coding the entity type to be TodoDB.class (we can safely ignore this exception)
             * 
             */
            Todo todo = em.find(TodoDB.class, id, LockModeType.PESSIMISTIC_WRITE);

            if (todo == null) {
                /* 
                 * If we could not find a Todo object, then we need to throw an exception.
                 * 
                 * Notice that we are inside a try/catch block, but since we do not
                 * catch this exception it will be thrown back to the caller (in this case the logic layer)
                 * 
                 */
                String message = "Could not find a Todo object with id: " + id;
                todoLogger.log(TodoLogger.TodoLoggerLevel.DEBUG, message, message);
                throw new TodoEntityNotFoundException(message);
            }

            todo.setActive(active);

            /* Similar reasoning as with persist in create method */
            em.merge(todo);

            /* Similar reasoning as in create method */
            em.getTransaction().commit();

        } catch (RollbackException e) {

            todoLogger.log(e);

            if (em.getTransaction().isActive()) {
                em.getTransaction().rollback();
            }

            throw new TodoServiceConfigurationException("Setting the Todo as active failed due to service errors. Please contact your database administrator.");

        } finally {
            em.close();
        }

    }

    @Override
    public Todo find(long id)
            throws
            TodoEntityNotFoundException,
            TodoServiceConfigurationException {

        EntityManager em = EMF.getEntityManager();

        try {

            Todo todo = em.find(TodoDB.class, id);
            if (todo == null) {
                String message = "No Todo for id: " + id;
                todoLogger.log(TodoLogger.TodoLoggerLevel.DEBUG, message, message);
                throw new TodoEntityNotFoundException(message);
            }

            return todo;

            /* 
             * There seems to be no exceptions thrown regarding issues with 
             * communicating with the database? This seems strange.
             * 
             * This is a case of when RuntimeExceptions might be coming from
             * further down the chain. It might be an MySQL exception or a 
             * Derby exception that is thrown, and you need to KNOW this
             * or go deeper into the documentation if you want to check for them.
             * 
             * If you do not check for them, and they happen, then they
             * will be thrown anyway (as they are runtime) and they will propagate
             * through the layers. The issue that might evolve from this is that
             * you have not logged the exception, so you do not know that 
             * something went wrong.
             * 
             */

        } finally {
            em.close();
        }

    }

    @Override
    public List<Todo> findAll()
            throws
            TodoServiceConfigurationException {

        EntityManager em = EMF.getEntityManager();

        try {

            /* createQuery might throw an exception if the query is 
             * syntactically incorrect. We will assume that it is not.
             */
            Query query = em.createQuery("SELECT t FROM TodoDB t");

            /*
             * getResultList has a lot of exceptions:
             * 
             * IllegalStateException - if called for a Java Persistence query language UPDATE or DELETE statement
             * QueryTimeoutException - if the query execution exceeds the query timeout value set and only the statement is rolled back
             * TransactionRequiredException - if a lock mode has been set and there is no transaction
             * PessimisticLockException - if pessimistic locking fails and the transaction is rolled back
             * LockTimeoutException - if pessimistic locking fails and only the statement is rolled back
             * PersistenceException - if the query execution exceeds the query timeout value set and the transaction is rolled back
             * 
             * The first exception will not happen if you have coded correctly.
             * 
             * The second exception might happen, and should be checked.
             * 
             * The third, fourth ,fifth and sixth exceptions are not going to happen here (i.e. you can control them).
             */
            List<Todo> results = query.getResultList();

            return results;
            
            /*
             * There are similar concerns here as in find, there might be exceptions
             * that come from deeper than the EntityManager...
             */

        } catch (QueryTimeoutException e) {

            todoLogger.log(e);
            throw new TodoServiceConfigurationException("A timout occured when communicating to the database, please contact your database administrator");

        } finally {
            em.close();
        }

    }
}
