package tdp024.todo.rest.service;

import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Response;
import tdp024.todo.data.api.entity.Todo;
import tdp024.todo.data.api.exception.TodoEntityNotFoundException;
import tdp024.todo.data.api.exception.TodoInputParameterException;
import tdp024.todo.data.api.exception.TodoServiceConfigurationException;
import tdp024.todo.data.db.facade.TodoEntityFacadeDB;
import tdp024.todo.logic.api.facade.TodoLogicFacade;
import tdp024.todo.logic.impl.facade.TodoLogicFacadeImpl;
import tdp024.util.json.TodoJsonSerializer;
import tdp024.util.json.TodoJsonSerializerImpl;
import tdp024.util.logger.TodoLogger;
import tdp024.util.logger.TodoLoggerImpl;

@Path("/todo")
public class TodoService {

    // --- Here we choose the implementations of the logic and data layer --- //
    private final TodoLogicFacade todoLogicFacde =
            new TodoLogicFacadeImpl(new TodoEntityFacadeDB());
    //----------------------------------------------------------------------- //
    private static final TodoLogger todoLogger = new TodoLoggerImpl();
    private static final TodoJsonSerializer jsonSerializer = new TodoJsonSerializerImpl();

    /*
     * 
     * Finally the exceptions reach the web layer.
     * Remebmer that there might be runtime exceptions that we have not taken
     * care of, and it might be a security issue presenting these to the outside
     * world.
     * 
     * There is something called an "exception mapper" in JAX-RS, but we will
     * take a more direct (but not as elegant) approach.
     * 
     */
    @GET
    @Path("create")
    public Response create(
            @QueryParam("title") String title,
            @QueryParam("body") String body) {

        try {

            long id = todoLogicFacde.create(title, body);

            /* If we get here then all is well, we have created a Todo */
            return Response.ok().entity(id + "").build();

        } catch (TodoInputParameterException e) {

            /* Something is wrong with the input parameters, let the caller know */
            todoLogger.log(e);
            return Response.status(Response.Status.BAD_REQUEST).entity("Something is wrong with your input").build();

        } catch (TodoServiceConfigurationException e) {

            /* Something is wrong internally, the caller is not able to fix this */
            todoLogger.log(e);
            return Response.status(Response.Status.INTERNAL_SERVER_ERROR).entity("The service could not handle your request, please contant your administrator").build();

        } catch (Throwable e) {

            /* 
             * Some other exception made it to the web layer, 
             * we do not want to leak information about this excpetion, but 
             * we MUST log it, so that we are made aware of this exception
             * that happened that we did not check for.
             * 
             */
            todoLogger.log(e);
            return Response.status(Response.Status.SERVICE_UNAVAILABLE).build();

        }




    }

    @GET
    @Path("find")
    public Response find(
            @QueryParam("id") long id) {

        /* Compare to the same method in the first version (without exception handling) */

        try {

            Todo todo = todoLogicFacde.find(id);

            /* If we get here then everything is fine, no need to check for todo being null */
            String json = jsonSerializer.toJson(todo);
            return Response.status(Response.Status.OK).entity(json).build();

        } catch (TodoEntityNotFoundException e) {

            todoLogger.log(e);
            return Response.status(Response.Status.BAD_REQUEST).entity("We could not find the todo you were looking for").build();

        } catch (TodoServiceConfigurationException e) {
            
            todoLogger.log(e);
            return Response.status(Response.Status.INTERNAL_SERVER_ERROR).entity("The service could not handle your request, please contant your administrator").build();
            
        } catch (Throwable e) {

            /*
             * Look at the reasoning in create method
             */
            todoLogger.log(e);
            return Response.status(Response.Status.SERVICE_UNAVAILABLE).build();
        }


    }
}
